# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Data models"""

import re
from typing import Optional

from metldata.builtin_transformations.infer_references.path.path_elements import (
    ReferencePathElement,
    ReferencePathElementType,
)

NAME_PATTERN = r"(?!\d)\w+"
ACTIVE_ARROW_PATTERN = rf"\({NAME_PATTERN}\)>"
PASSIVE_ARROW_PATTERN = rf"<\({NAME_PATTERN}\)"
ARROW_PATTERN = rf"(({ACTIVE_ARROW_PATTERN})|({PASSIVE_ARROW_PATTERN}))"
ELEMENT_PATTERN = rf"{NAME_PATTERN}{ARROW_PATTERN}{NAME_PATTERN}"
PATH_RAW_CHAR_PATTERN = r"^[\w><\(\)]+$"
PATH_PATTERN = (
    rf"^{NAME_PATTERN}{ARROW_PATTERN}({NAME_PATTERN}{ARROW_PATTERN})*{NAME_PATTERN}$"
)


class ValidationError(RuntimeError):
    """Raised when a path string was invalid"""


def clean_path_str(path_str: str) -> str:
    """Cleanup whitespaces, newlines, etc."""

    return "".join(path_str.split())


def validate_path_str_characters(path_str: str) -> None:
    """Validates the characters of the an uncleaned path str. The path_str is assumed to
    be cleaned.

    Raises:
        ValidationError: if invalid.
    """

    if not re.match(PATH_RAW_CHAR_PATTERN, path_str):
        raise ValidationError(
            f"The following path string contains invalid characters: {path_str}"
        )


def validate_path_str_format(path_str: str) -> None:
    """Validates the format of the path str. The path_str is assumed to be cleaned.

    Raises:
        ValidationError: if invalid.
    """

    if not re.match(PATH_PATTERN, path_str):
        raise ValidationError(
            f"The following path string has an invalid format: {path_str}"
        )


def validate_string_element(string_element: str) -> None:
    """Validates the format of a string-based path element. The path_str is assumed to
    be cleaned.

    Raises:
        ValidationError: if invalid.
    """

    if not re.match(rf"^{ELEMENT_PATTERN}$", string_element):
        raise ValidationError(
            "The following string-based path element has an invalid format: "
            + string_element
        )


def extract_first_element(*, path_str: str) -> str:
    """Extract the first element of a path_str. The path_str is assumed to be cleaned.

    Raises:
        ValidationError: if no element can be extracted.
    """

    match = re.match(rf"^({ELEMENT_PATTERN}).*$", path_str)

    if not match:
        raise ValidationError(f"Cannot find element in path string: {path_str}")

    return match.group(1)


def get_target_class(*, path_str: str) -> str:
    """Get the target class of a path str. The path_str is assumed to be cleaned."""

    match = re.match(rf"^.*?({NAME_PATTERN})$", path_str)

    if not match:
        raise ValidationError(f"Cannot find target class of path string: {path_str}")

    return match.group(1)


def split_first_element(*, path_str: str) -> tuple[str, Optional[str]]:
    """Return a tuple of the first element and the remaining path string.
    Thereby, the target class of the first element is set as the source class of the
    remaining path.
    The second element is None if the provided path only contained one element.
    The path_str is assumed to be cleaned.
    """

    first_element = extract_first_element(path_str=path_str)
    first_element_target_class = get_target_class(path_str=first_element)

    if first_element == path_str:
        return first_element, None

    remaining_path = path_str[len(first_element) :]
    remaining_path_extended = first_element_target_class + remaining_path

    return first_element, remaining_path_extended


def get_string_elements(*, path_str: str) -> list[str]:
    """Decomposes a path string into elements in string repesentation. The path_str is
    assumed to be cleaned."""

    elements: list[str] = []
    remaining_path = path_str

    # extract one element at a time:
    while remaining_path:
        element, remaining_path = split_first_element(  # type: ignore
            path_str=remaining_path
        )
        elements.append(element)

    return elements


def get_element_type(*, string_element: str) -> ReferencePathElementType:
    """Infers the type of the provided string-based element."""

    validate_string_element(string_element)

    return (
        ReferencePathElementType.ACTIVE
        if ">" in string_element
        else ReferencePathElementType.PASSIVE
    )


def get_element_components(*, string_element: str) -> tuple[str, str, str]:
    """Returns a tuple of the source, the slot, and the target of the string-based path
    element."""

    # remove the angle:
    string_element_cleaned = string_element.replace(">", "").replace("<", "")

    # extract the source:
    source, slot_and_target = string_element_cleaned.split("(")

    # extract slot and target:
    slot, target = slot_and_target.split(")")

    return source, slot, target


def string_element_to_object(string_element: str) -> ReferencePathElement:
    """Translates a string-based path element into an object-based representation."""

    validate_string_element(string_element)
    type_ = get_element_type(string_element=string_element)
    source, slot, target = get_element_components(string_element=string_element)

    return ReferencePathElement(type_=type_, source=source, slot=slot, target=target)


def path_str_to_object_elements(path_str: str) -> list[ReferencePathElement]:
    """Translates a path string into a list of object-based elements. The path_str is
    assumed to be cleaned."""

    validate_path_str_characters(path_str=path_str)
    validate_path_str_format(path_str=path_str)

    string_elements = get_string_elements(path_str=path_str)
    return [
        string_element_to_object(string_element) for string_element in string_elements
    ]



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test reference utils."""

from contextlib import nullcontext

import pytest
from pydantic import BaseModel

from metldata.builtin_transformations.infer_references.path.path import ReferencePath
from metldata.builtin_transformations.infer_references.path.path_elements import (
    ReferencePathElement,
    ReferencePathElementType,
)


@pytest.mark.parametrize(
    "path_str, expected_elements, expected_source, expected_target",
    [
        (
            "class_a(class_b)>class_b",
            [
                ReferencePathElement(
                    type_=ReferencePathElementType.ACTIVE,
                    source="class_a",
                    slot="class_b",
                    target="class_b",
                )
            ],
            "class_a",
            "class_b",
        ),
        (
            """class_a
                (class_b) >
                class_b""",  # containing whitespaces
            [
                ReferencePathElement(
                    type_=ReferencePathElementType.ACTIVE,
                    source="class_a",
                    slot="class_b",
                    target="class_b",
                )
            ],
            "class_a",
            "class_b",
        ),
        (
            "class_a<(class_a)class_b",
            [
                ReferencePathElement(
                    type_=ReferencePathElementType.PASSIVE,
                    source="class_a",
                    slot="class_a",
                    target="class_b",
                )
            ],
            "class_a",
            "class_b",
        ),
        (
            "class_a(class_b)>class_b(class_c)>class_c",
            [
                ReferencePathElement(
                    type_=ReferencePathElementType.ACTIVE,
                    source="class_a",
                    slot="class_b",
                    target="class_b",
                ),
                ReferencePathElement(
                    type_=ReferencePathElementType.ACTIVE,
                    source="class_b",
                    slot="class_c",
                    target="class_c",
                ),
            ],
            "class_a",
            "class_c",
        ),
        (
            "class_a(class_b)>class_b<(class_b)class_c",
            [
                ReferencePathElement(
                    type_=ReferencePathElementType.ACTIVE,
                    source="class_a",
                    slot="class_b",
                    target="class_b",
                ),
                ReferencePathElement(
                    type_=ReferencePathElementType.PASSIVE,
                    source="class_b",
                    slot="class_b",
                    target="class_c",
                ),
            ],
            "class_a",
            "class_c",
        ),
    ],
)
def test_reference_path(
    path_str: str,
    expected_elements: ReferencePathElement,
    expected_source: str,
    expected_target: str,
):
    """Test the ReferencePath class."""

    observed_path = ReferencePath(path_str=path_str)
    assert observed_path.elements == expected_elements
    assert observed_path.source == expected_source
    assert observed_path.target == expected_target


@pytest.mark.parametrize(
    "path_str, is_valid",
    [
        ("class_a(class_b)>class_b", True),
        ("class_a<(class_a)class_b", True),
        ("class_a(class_b)>class_b(class_c)>class_c", True),
        (12312, False),
        ("class_a<(class_b)>class_b", False),
        ("(class_b)>class_b(class_c)>class_c", False),
    ],
)
def test_reference_path_pydantic(path_str: str, is_valid: bool):
    """Test the ReferencePath class when used with pydantic."""

    class ExampleModel(BaseModel):
        """Some example model."""

        path: ReferencePath

    with nullcontext() if is_valid else pytest.raises(ValueError):  # type: ignore
        observed_path = ExampleModel(path=path_str).path

    if is_valid:
        expected_path = ReferencePath(path_str=path_str)
        assert observed_path == expected_path



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the anchor module."""


import pytest

from metldata.model_utils.anchors import (
    AnchorPoint,
    AnchorPointNotFoundError,
    InvalidAnchorPointError,
    filter_anchor_points,
    get_anchor_points,
    get_anchors_points_by_target,
    lookup_anchor_point,
)
from metldata.model_utils.essentials import MetadataModel
from tests.fixtures.metadata_models import (
    ANCHORS_INVALID_MODELS,
    VALID_MINIMAL_METADATA_MODEL,
)

EXAMPLE_ANCHOR_POINTS = {
    AnchorPoint(target_class="File", identifier_slot="alias", root_slot="files"),
    AnchorPoint(target_class="Dataset", identifier_slot="alias", root_slot="datasets"),
}

EXAMPLE_ANCHOR_POINTS_BY_TARGET = {
    "File": AnchorPoint(
        target_class="File", identifier_slot="alias", root_slot="files"
    ),
    "Dataset": AnchorPoint(
        target_class="Dataset", identifier_slot="alias", root_slot="datasets"
    ),
}


def test_get_anchor_points_happy():
    """Test the happy path of using the get_anchor_points function."""

    expected_anchor_points = EXAMPLE_ANCHOR_POINTS

    observed_anchor_points = get_anchor_points(model=VALID_MINIMAL_METADATA_MODEL)

    assert observed_anchor_points == expected_anchor_points


@pytest.mark.parametrize("invalid_model", ANCHORS_INVALID_MODELS)
def test_get_anchor_points_invalid(invalid_model: MetadataModel):
    """Test the get_anchor_points function for models with invalid anchor points."""

    with pytest.raises(InvalidAnchorPointError):
        _ = get_anchor_points(model=invalid_model)


def test_get_anchor_points_by_target_happy():
    """Test the happy path of using the get_anchors_by_target function."""

    expected_anchor_points = EXAMPLE_ANCHOR_POINTS_BY_TARGET

    observed_anchor_points = get_anchors_points_by_target(
        model=VALID_MINIMAL_METADATA_MODEL
    )

    assert observed_anchor_points == expected_anchor_points


def test_filter_anchor_points_happy():
    """Test the happy path of using the filter_anchor_points function."""

    class_of_interest = "File"
    expected_anchor_points = {
        class_of_interest: EXAMPLE_ANCHOR_POINTS_BY_TARGET[class_of_interest],
    }

    observed_anchor_points = filter_anchor_points(
        anchor_points_by_target=EXAMPLE_ANCHOR_POINTS_BY_TARGET,
        classes_of_interest={class_of_interest},
    )

    assert observed_anchor_points == expected_anchor_points


def test_filter_anchor_points_non_existing_class():
    """Test the happy path of using the filter_anchor_points function."""

    classes_of_interest = set(EXAMPLE_ANCHOR_POINTS_BY_TARGET.keys())
    classes_of_interest.add("NonExisting")

    with pytest.raises(AnchorPointNotFoundError):
        _ = filter_anchor_points(
            anchor_points_by_target=EXAMPLE_ANCHOR_POINTS_BY_TARGET,
            classes_of_interest=classes_of_interest,
        )


def test_lookup_anchor_point_happy():
    """Test the happy path of using the lookup_anchor_point function."""

    class_name = "File"
    expected_anchor_point = EXAMPLE_ANCHOR_POINTS_BY_TARGET[class_name]

    observed_anchor_point = lookup_anchor_point(
        class_name=class_name, anchor_points_by_target=EXAMPLE_ANCHOR_POINTS_BY_TARGET
    )

    assert observed_anchor_point == expected_anchor_point


def test_lookup_anchor_point_not_found():
    """Test the using the lookup_anchor_point function with a class name that does
    not exist."""

    with pytest.raises(AnchorPointNotFoundError):
        _ = lookup_anchor_point(
            class_name="NonExisting",
            anchor_points_by_target=EXAMPLE_ANCHOR_POINTS_BY_TARGET,
        )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for transforming metadata."""

from collections import defaultdict

from pydantic import Json
from typing_extensions import TypeAlias

from metldata.metadata_utils import lookup_self_id
from metldata.model_utils.anchors import AnchorPoint, lookup_anchor_point
from metldata.model_utils.essentials import MetadataModel
from metldata.submission_registry.models import AccessionMap
from metldata.transform.base import MetadataTransformationError

# A type for specifying references between anchored classes in a given model:
# The first key is the name of the class that is anchored, the second key is the
# the name of the slot of that class that contains the references.
# The value is name of the class that is referenced.
References: TypeAlias = dict[str, dict[str, str]]


def get_references(
    *, metadata_model: MetadataModel, anchor_points_by_target: dict[str, AnchorPoint]
) -> References:
    """Get references between anchored classes in a given model."""

    schema_view = metadata_model.schema_view
    references: References = defaultdict(dict)

    for source_class_name in anchor_points_by_target:
        slots = schema_view.class_induced_slots(source_class_name)
        for slot in slots:
            if slot.range in anchor_points_by_target:
                references[source_class_name][slot.name] = slot.range

    return references


def lookup_accession(
    *,
    target_class: str,
    old_identifier: str,
    accession_map: AccessionMap,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> str:
    """Lookup the accession for the a resource with the given identifier of the given
    class."""
    anchor_point = lookup_anchor_point(
        class_name=target_class, anchor_points_by_target=anchor_points_by_target
    )
    accession = accession_map.get(anchor_point.root_slot, {}).get(old_identifier)
    if not accession:
        raise MetadataTransformationError(
            f"Could not find accession for '{old_identifier}' of class"
            + f" '{target_class}."
        )
    return accession


def add_accession_to_resource(
    *,
    resource: Json,
    class_name: str,
    old_identifier_slot: str,
    accession_slot_name: str,
    accession_map: AccessionMap,
    references: dict[str, str],
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Add an accession to a resource.

    Args:
        resource:
            The resource to which accessions should be added.
        old_identifier_slot:
            The name of the slot that contains the old identifier.
        old_identifier:
            The old identifier of the resource.
        accession_map:
            The accession map that contains the accessions.
        references:
            References from this to other anchored classes. The keys are the names of
            the slots that contain the references, the values are the names of the
            referenced classes.

    Raises:
        MetadataTransformationError:
            if the transformation of the metadata fails.
    """

    old_identifier = lookup_self_id(
        resource=resource, identifier_slot=old_identifier_slot
    )
    new_identifier = lookup_accession(
        target_class=class_name,
        old_identifier=old_identifier,
        accession_map=accession_map,
        anchor_points_by_target=anchor_points_by_target,
    )

    new_resource: Json = {accession_slot_name: new_identifier}

    for slot_name, slot_value in resource.items():
        if slot_name in references:
            referenced_class = references[slot_name]
            if isinstance(slot_value, list):
                new_resource[slot_name] = [
                    lookup_accession(
                        target_class=referenced_class,
                        old_identifier=reference_old_identifier,
                        accession_map=accession_map,
                        anchor_points_by_target=anchor_points_by_target,
                    )
                    for reference_old_identifier in slot_value
                ]
            else:
                new_resource[slot_name] = lookup_accession(
                    target_class=referenced_class,
                    old_identifier=slot_value,
                    accession_map=accession_map,
                    anchor_points_by_target=anchor_points_by_target,
                )
        else:
            new_resource[slot_name] = slot_value

    return new_resource


def add_accessions_to_metadata(
    *,
    metadata: Json,
    accession_slot_name: str,
    accession_map: AccessionMap,
    references: References,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Add an accessions to metadata.

    Raises:
        MetadataTransformationError:
            if the transformation of the metadata fails.
    """

    modified_metadata = {}

    for target_class_name, anchor_point in anchor_points_by_target.items():
        target_resources = []
        target_accession_map = accession_map.get(anchor_point.root_slot)
        if target_accession_map is None:
            raise MetadataTransformationError(
                "Could not find accession mapping for target class"
                + f" {target_class_name}."
            )

        for old_resource in metadata[anchor_point.root_slot]:
            new_resource = add_accession_to_resource(
                resource=old_resource,
                class_name=target_class_name,
                old_identifier_slot=anchor_point.identifier_slot,
                accession_slot_name=accession_slot_name,
                accession_map=accession_map,
                references=references[target_class_name],
                anchor_points_by_target=anchor_points_by_target,
            )
            target_resources.append(new_resource)

        modified_metadata[anchor_point.root_slot] = target_resources

    return modified_metadata



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testing the submission store."""

import pytest
from ghga_service_commons.utils.utc_dates import now_as_utc

from metldata.config import SubmissionConfig
from metldata.submission_registry.models import (
    StatusChange,
    Submission,
    SubmissionStatus,
)
from metldata.submission_registry.submission_store import SubmissionStore
from tests.fixtures.config import config_sub_fixture  # noqa: F401

EXAMPLE_SUBMISSION = Submission(
    title="test",
    description="test",
    content={"test_class": [{"alias": "test_alias1"}]},
    accession_map={"test_class": {"test_alias1": "test_accession1"}},
    id="testsubmission001",
    status_history=(
        StatusChange(
            timestamp=now_as_utc(),
            new_status=SubmissionStatus.COMPLETED,
        ),
    ),
)


def test_happy(config_sub_fixture: SubmissionConfig):  # noqa: F811
    """Test the happy path of inserting, querying, and updating a submission using
    the submission store."""

    submission_store = SubmissionStore(config=config_sub_fixture)

    # make sure that sumbission does not exist:
    assert not submission_store.exists(submission_id=EXAMPLE_SUBMISSION.id)

    # create new submission:
    submission_store.insert_new(submission=EXAMPLE_SUBMISSION)

    # make sure that sumbission exists:
    assert submission_store.exists(submission_id=EXAMPLE_SUBMISSION.id)

    # query newly created submission:
    submission_queried = submission_store.get_by_id(submission_id=EXAMPLE_SUBMISSION.id)
    assert EXAMPLE_SUBMISSION == submission_queried

    # update the submision:
    submission_updated = EXAMPLE_SUBMISSION.copy(
        update={"title": "updated test submission"}
    )
    submission_store.update_existing(submission=submission_updated)

    # query updated submission:
    submission_updated_queried = submission_store.get_by_id(
        submission_id=submission_updated.id
    )
    assert submission_updated_queried == submission_updated


def test_query_non_existing(
    config_sub_fixture: SubmissionConfig,  # noqa: F811
):
    """Test querying for a non-existing submission."""

    submission_store = SubmissionStore(config=config_sub_fixture)

    with pytest.raises(SubmissionStore.SubmissionDoesNotExistError):
        _ = submission_store.get_by_id(submission_id="non-exisitng-id")



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the builtin workflows using pre-defined test cases."""

import pytest

from metldata.transform.handling import WorkflowHandler
from tests.fixtures.workflows import WORKFLOW_TEST_CASES, WorkflowTestCase


@pytest.mark.parametrize(
    "test_case",
    WORKFLOW_TEST_CASES,
    ids=str,
)
def test_model_transform(
    test_case: WorkflowTestCase,
):
    """Test the happy path of transforming a model."""

    handler = WorkflowHandler(
        workflow_definition=test_case.workflow_definition,
        workflow_config=test_case.config,
        original_model=test_case.original_model,
    )
    artifact_models = handler.artifact_models

    for artifact, expected_model in test_case.artifact_models.items():
        assert artifact_models[artifact] == expected_model


@pytest.mark.parametrize("test_case", WORKFLOW_TEST_CASES, ids=str)
def test_metadata_transform(
    test_case: WorkflowTestCase,
):
    """Test the happy path of transforming metadata for a model."""

    handler = WorkflowHandler(
        workflow_definition=test_case.workflow_definition,
        workflow_config=test_case.config,
        original_model=test_case.original_model,
    )
    artifact_metadata = handler.run(
        metadata=test_case.original_metadata, annotation=test_case.submission_annotation
    )

    for artifact, expected_metadata in test_case.artifact_metadata.items():
        assert artifact_metadata[artifact] == expected_metadata



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""The CLI interface."""

import asyncio
import functools

import typer
from ghga_service_commons.api import run_server

from metldata.combined import get_app
from metldata.config import Config

cli = typer.Typer()


def run_sync(coroutine):
    """A decorator to run a coroutine as a synchronous function."""

    @functools.wraps(coroutine)
    def wrapper(*args, **kwargs):
        return asyncio.run(coroutine(*args, **kwargs))

    return wrapper


@cli.command()
@run_sync
async def run_api() -> None:
    """Run the combined loader and query API."""

    config = Config()
    app = await get_app(config=config)
    await run_server(app=app, config=config)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the artifact_info module."""

from metldata.artifacts_rest.artifact_info import (
    load_artifact_info,
    subset_json_schema_for_class,
)
from tests.fixtures.metadata_models import VALID_MINIMAL_METADATA_MODEL


def test_subset_json_schema_for_class():
    """Test happy path of using subset_json_schema_for_class."""

    # a global schema that uses File and Dataset definitions on the top level:
    global_json_schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "properties": {
            "datasets": {
                "items": {
                    "additionalProperties": {"$ref": "#/$defs/Dataset"},
                    "type": "object",
                },
                "type": "array",
            },
            "files": {
                "items": {
                    "additionalProperties": {"$ref": "#/$defs/File"},
                    "type": "object",
                },
                "type": "array",
            },
        },
        "required": ["files", "datasets"],
        "type": "object",
        "$defs": {
            "Dataset": {
                "properties": {
                    "files": {
                        "items": {
                            "additionalProperties": {"$ref": "#/$defs/File"},
                            "type": "object",
                        },
                        "type": "array",
                    }
                },
                "required": ["files"],
                "type": "object",
            },
            "File": {
                "properties": {
                    "checksum": {"type": "string"},
                    "size": {"type": "integer"},
                },
                "required": ["size", "checksum"],
                "type": "object",
            },
        },
    }

    # generate a subset schema for the Dataset class:
    observed_subschema = subset_json_schema_for_class(
        global_json_schema=global_json_schema, class_name="Dataset"
    )

    # the subset schema should only contain the File defintion as it is used by the
    # Dataset class:
    expected_subschema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "properties": {
            "files": {
                "items": {
                    "additionalProperties": {"$ref": "#/$defs/File"},
                    "type": "object",
                },
                "type": "array",
            }
        },
        "required": ["files"],
        "type": "object",
        "$defs": {
            "File": {
                "properties": {
                    "checksum": {"type": "string"},
                    "size": {"type": "integer"},
                },
                "required": ["size", "checksum"],
                "type": "object",
            },
        },
    }
    assert observed_subschema == expected_subschema


def test_load_artifact_info():
    """Test happy path of using load_artifact_info."""

    expected_artifact_name = "test_artifact"
    expected_artifact_description = "This is a test artifact."
    expected_resource_class_names = {"File", "Dataset"}

    artifact_info = load_artifact_info(
        model=VALID_MINIMAL_METADATA_MODEL,
        name=expected_artifact_name,
        description=expected_artifact_description,
    )

    assert artifact_info.name == expected_artifact_name
    assert artifact_info.description == expected_artifact_description
    assert artifact_info.resource_classes.keys() == expected_resource_class_names



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the handling module. Only edge cases that are not covered by tests
with builtin transformations are tested here."""

import pytest

from metldata.builtin_transformations.infer_references.main import (
    REFERENCE_INFERENCE_TRANSFORMATION,
    ReferenceInferenceConfig,
)
from metldata.model_utils.essentials import MetadataModel
from metldata.transform.base import (
    MetadataModelAssumptionError,
    MetadataModelTransformationError,
    TransformationDefinition,
)
from metldata.transform.handling import TransformationHandler
from tests.fixtures.metadata_models import VALID_ADVANCED_METADATA_MODEL

VALID_EXAMPLE_CONFIG = ReferenceInferenceConfig(
    inferred_ref_map={
        "Experiment": {
            "files": {
                "path": "Experiment(samples)>Sample(files)>File",
                "multivalued": True,
            }
        }
    }
)


def test_transformation_handler_assumption_error():
    """Test using the TransformationHandling when model assumptions are not met."""

    # make transformation definition always raise an MetadataModelAssumptionError:
    def always_failing_assumptions(
        model: MetadataModel, config: ReferenceInferenceConfig
    ):
        """A function that always raises a MetadataModelAssumptionError."""
        raise MetadataModelAssumptionError

    transformation = TransformationDefinition(
        config_cls=REFERENCE_INFERENCE_TRANSFORMATION.config_cls,
        check_model_assumptions=always_failing_assumptions,
        transform_model=REFERENCE_INFERENCE_TRANSFORMATION.transform_model,
        metadata_transformer_factory=REFERENCE_INFERENCE_TRANSFORMATION.metadata_transformer_factory,
    )

    with pytest.raises(MetadataModelAssumptionError):
        _ = TransformationHandler(
            transformation_definition=transformation,
            transformation_config=VALID_EXAMPLE_CONFIG,
            original_model=VALID_ADVANCED_METADATA_MODEL,
        )


def test_transformation_handler_model_transformation_error():
    """Test using the TransformationHandling when model transformation fails."""

    # make transformation definition always raise an MetadataModelAssumptionError:
    def always_failing_transformation(
        original_model: MetadataModel, config: ReferenceInferenceConfig
    ):
        """A function that always raises a MetadataModelTransformationError."""
        raise MetadataModelTransformationError

    transformation = TransformationDefinition(
        config_cls=REFERENCE_INFERENCE_TRANSFORMATION.config_cls,
        check_model_assumptions=REFERENCE_INFERENCE_TRANSFORMATION.check_model_assumptions,
        transform_model=always_failing_transformation,
        metadata_transformer_factory=REFERENCE_INFERENCE_TRANSFORMATION.metadata_transformer_factory,
    )

    with pytest.raises(MetadataModelTransformationError):
        _ = TransformationHandler(
            transformation_definition=transformation,
            transformation_config=VALID_EXAMPLE_CONFIG,
            original_model=VALID_ADVANCED_METADATA_MODEL,
        )
